==> irreco-core_0.6.2.tar.gz: irreco-core-0.6.2/src/core/irreco_cmd_chain.c <==
/*
 * irreco - Ir Remote Control
 * Copyright (C) 2007  Arto Karppinen (arto.karppinen@iki.fi)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "irreco_cmd_chain.h"
#include <hildon/hildon-banner.h>

/**
 * @typedef IrrecoCmdChain
 *
 * IrrecoCmdChain contains a list of IrrecoCmds, which can then all be 
 * executed one after another by calling irreco_cmd_chain_execute().
 */
 
 
 


 
IrrecoCmdChain *irreco_cmd_chain_create()
{
	IrrecoCmdChain *irreco_cmd_chain;
	IRRECO_ENTER
	
	irreco_cmd_chain = g_slice_new0(IrrecoCmdChain);
	irreco_cmd_chain->execution_rate = IRRECO_SECONDS_TO_USEC(0.3);
	IRRECO_RETURN_PTR(irreco_cmd_chain);
}

void irreco_cmd_chain_remove_all(IrrecoCmdChain * irreco_cmd_chain)
{
	IRRECO_ENTER
	
	g_assert(irreco_cmd_chain != NULL);
	
	while (irreco_cmd_chain->command_list != NULL) {
		irreco_cmd_destroy(irreco_cmd_chain->command_list->data);
		irreco_cmd_chain->command_list = g_list_remove(
			irreco_cmd_chain->command_list,
			irreco_cmd_chain->command_list->data);
	}
	IRRECO_RETURN
}

void irreco_cmd_chain_destroy(IrrecoCmdChain * irreco_cmd_chain)
{
	IRRECO_ENTER
	if (irreco_cmd_chain == NULL) IRRECO_RETURN
	irreco_cmd_chain_remove_all(irreco_cmd_chain);
	g_slice_free(IrrecoCmdChain, irreco_cmd_chain);
	IRRECO_RETURN
}

void irreco_cmd_chain_set_execution_rate(IrrecoCmdChain * irreco_cmd_chain, 
					 glong execution_rate)
{
	IRRECO_ENTER
	g_assert(irreco_cmd_chain != NULL);
	IRRECO_PRINTF("Setting execution rate to \"%li\".\n", execution_rate);
	irreco_cmd_chain->execution_rate = execution_rate;
	IRRECO_RETURN
}

void irreco_cmd_chain_append(IrrecoCmdChain * irreco_cmd_chain, 
			     IrrecoCmd * irreco_cmd)
{
	IRRECO_ENTER
	
	g_assert(irreco_cmd_chain != NULL);
	
	irreco_cmd_chain->command_list = g_list_append(
		g_list_last(irreco_cmd_chain->command_list), irreco_cmd);
	IRRECO_RETURN
}

void irreco_cmd_chain_append_copy(IrrecoCmdChain * irreco_cmd_chain, 
				  IrrecoCmd * irreco_cmd)
{
	IRRECO_ENTER
	
	g_assert(irreco_cmd_chain != NULL);
	
	irreco_cmd_chain->command_list = g_list_append(
		g_list_last(irreco_cmd_chain->command_list),
		irreco_cmd_dublicate(irreco_cmd));
	IRRECO_RETURN
}

void irreco_cmd_chain_remove(IrrecoCmdChain * irreco_cmd_chain, guint index)
{
	IrrecoCmd *irreco_cmd;
	IRRECO_ENTER
	
	g_assert(irreco_cmd_chain != NULL);
	
	IRRECO_DEBUG("Removing command from index \"%u\".\n", index);
	irreco_cmd = (IrrecoCmd *) g_list_nth_data(g_list_first(
		irreco_cmd_chain->command_list), index);
	irreco_cmd_destroy(irreco_cmd);
	irreco_cmd_chain->command_list = g_list_remove(
		g_list_first(irreco_cmd_chain->command_list), irreco_cmd);

	IRRECO_RETURN
}

void irreco_cmd_chain_move(IrrecoCmdChain * irreco_cmd_chain, 
			   guint from_index, guint to_index)
{
	gpointer data;
	GList* from_link;
	gint length;
	IRRECO_ENTER
	
	length = irreco_cmd_chain_length(irreco_cmd_chain);
	
	if (from_index < 0) {
		from_index = 0;
	} else if (from_index > length) {
		IRRECO_RETURN
	}
	
	if (to_index > length) {
		to_index = length;
	} else if (to_index < 0) {
		IRRECO_RETURN
	}
	
	from_link = g_list_nth(g_list_first(
		irreco_cmd_chain->command_list), from_index);
	data = from_link->data;
	irreco_cmd_chain->command_list = g_list_delete_link(
		irreco_cmd_chain->command_list, from_link);
	irreco_cmd_chain->command_list = g_list_insert(g_list_first(
		irreco_cmd_chain->command_list), data, to_index);

	IRRECO_RETURN
}

/**
 * Make a copy of a command chain.
 *
 * @param from  Make a copy of this IrrecoCmdChain, or NULL.
 */
void irreco_cmd_chain_copy(IrrecoCmdChain * from, 
			   IrrecoCmdChain * to)
{
	IRRECO_ENTER
	irreco_cmd_chain_remove_all(to);
	to->execution_rate = from->execution_rate;
	
	if (from == NULL) IRRECO_RETURN
	IRRECO_CMD_CHAIN_FOREACH(from, command)
		irreco_cmd_chain_append_copy(to, command);
	IRRECO_CMD_CHAIN_FOREACH_END

	IRRECO_RETURN
}

guint irreco_cmd_chain_length(IrrecoCmdChain * irreco_cmd_chain)
{
	IRRECO_ENTER
	IRRECO_RETURN_UINT(g_list_length(g_list_first(
		irreco_cmd_chain->command_list)));
}

guint irreco_cmd_chain_lenght(IrrecoCmdChain * irreco_cmd_chain)
{
	IRRECO_ENTER
	if (irreco_cmd_chain->command_list == NULL) IRRECO_RETURN_UINT(0);
	IRRECO_RETURN_UINT(g_list_length(g_list_first(
			   irreco_cmd_chain->command_list)));
}

IrrecoCmd *irreco_cmd_chain_get(IrrecoCmdChain * irreco_cmd_chain,
				guint from_index)
{
	IRRECO_ENTER
	if (irreco_cmd_chain->command_list == NULL) IRRECO_RETURN_PTR(NULL);
	IRRECO_RETURN_PTR(g_list_nth_data(g_list_first(
			  irreco_cmd_chain->command_list), 
			  from_index));
}

void irreco_cmd_chain_print(IrrecoCmdChain * irreco_cmd_chain)
{
	gint i = 0;
	guint len;
	IRRECO_ENTER
	
	if (irreco_cmd_chain == NULL) {
		IRRECO_ERROR("IrrecoCmdChain pointer is NULL.\n");
		IRRECO_RETURN
	}
	
	len = irreco_cmd_chain_lenght(irreco_cmd_chain);
	IRRECO_PRINTF("Printing command chain. Lenght: %u\n", len);
	IRRECO_CMD_CHAIN_FOREACH(irreco_cmd_chain, command)
		IRRECO_PRINTF("Command %i / %u\n", ++i, len);
		irreco_cmd_print(command);
	IRRECO_CMD_CHAIN_FOREACH_END

	IRRECO_RETURN
}



/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/* Command chain execution.                                                   */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

typedef struct _IrrecoCmdChainExecute IrrecoCmdChainExecute;
struct _IrrecoCmdChainExecute {
	
	IrrecoData     *irreco_data;
	GFunc           call_when_done;
	gpointer        user_data;
	
	GString        *banner_message;
	GtkWidget      *banner;
	
	/* Execution time will be timed in relation of time_sync. */
	GTimeVal        time_sync;
	glong		time_pos;
	
	IrrecoCmdChain *chain;
	IrrecoCmd      *command;
	guint           pos;
	guint           len;
};

gboolean irreco_cmd_chain_execute_next_1(IrrecoCmdChainExecute *execute);
gboolean irreco_cmd_chain_execute_next_2(IrrecoCmdChainExecute *execute);

/**
 * Start executing the command chain.
 *
 * If the command chain is longer than one command, then delayed execution is 
 * started. In that case, call_when_done will be called when the command chain 
 * has finished executing.
 *
 * Data pointer will be the command chain, and user_data what user has given.
 *
 * @return TRUE if delayed execution started, FALSE otherwise.
 */
gboolean irreco_cmd_chain_execute(IrrecoCmdChain * irreco_cmd_chain,
				  IrrecoData *irreco_data,
				  GFunc call_when_done,
				  gpointer user_data)
{
	guint len;
	IRRECO_ENTER
	
	/* Empty list. */
	if (irreco_cmd_chain->command_list == NULL) {
		IRRECO_RETURN_BOOL(FALSE);
		
	/* Only one comamnd. */
	} else if ((len = irreco_cmd_chain_length(irreco_cmd_chain)) == 1) {
		irreco_cmd_execute(irreco_cmd_chain_get(irreco_cmd_chain, 0),
				   irreco_data);
		IRRECO_RETURN_BOOL(FALSE);
			
	/* Multiple commands. */
	} else {
		IrrecoCmdChainExecute *execute;
		
		execute = g_slice_new0(IrrecoCmdChainExecute);
		execute->banner_message = g_string_new("");
		execute->irreco_data = irreco_data;
		execute->chain       = irreco_cmd_chain;
		execute->pos         = 0;
		execute->len         = len;
		
		execute->call_when_done	= call_when_done;
		execute->user_data = user_data;
		
		g_get_current_time(&execute->time_sync);
		irreco_cmd_chain_execute_next_1(execute);
	}
	IRRECO_RETURN_BOOL(TRUE);
}

/**
 * Command execution phaze one, update banner.
 */
gboolean irreco_cmd_chain_execute_next_1(IrrecoCmdChainExecute *execute)
{
	gdouble fraction;
	IRRECO_ENTER
	
	IRRECO_PRINTF("Command %i / %u\n", execute->pos + 1, execute->len);
	execute->command = irreco_cmd_chain_get(execute->chain, execute->pos);
	
	g_string_printf(execute->banner_message, "%s%s", _("Executing: "),
			irreco_cmd_get_long_name(execute->command));
	IRRECO_PRINTF("%s\n", execute->banner_message->str);
	
	if (execute->banner == NULL) {
		GtkWidget *owner;
		owner = GTK_WIDGET(irreco_window_manager_get_gtk_window(
				   execute->irreco_data->window_manager));
		execute->banner	= hildon_banner_show_progress(
			owner, NULL, execute->banner_message->str);
	} else {
		hildon_banner_set_text(HILDON_BANNER(execute->banner), 
				       execute->banner_message->str);
	}
			       
	fraction = (float) 1 / (float) execute->len * (float)(execute->pos + 1);
	IRRECO_DEBUG("Proggressbar fraction %f\n", fraction);
	hildon_banner_set_fraction(HILDON_BANNER(execute->banner), fraction);
	g_idle_add(G_SOURCEFUNC(irreco_cmd_chain_execute_next_2), execute);
	IRRECO_RETURN_BOOL(FALSE);
}

/**
 * Command execution phaze two, execute command.
 */
gboolean irreco_cmd_chain_execute_next_2(IrrecoCmdChainExecute *execute)
{
	gboolean result;
	IRRECO_ENTER
	
	/* Wait command delay is simply added to the time_pos. */
	if (execute->command->type == IRRECO_COMMAND_WAIT) {
		
		/* Wait command overrides the chain specific execution rate.*/
		execute->time_pos = execute->time_pos
				    + execute->command->wait.delay;
		execute->pos = execute->pos + 1;
		result = TRUE;
	} else {
		result = irreco_cmd_execute(execute->command, 
					    execute->irreco_data);
		execute->time_pos = execute->time_pos
				    + execute->chain->execution_rate;
		execute->pos = execute->pos + 1;
	}
	
	/* Is something remaining to be executed? */
	if (result == TRUE && execute->pos < execute->len) {
		
		GTimeVal time_current, time_next;
		glong time_diff;
		
		g_get_current_time(&time_current);
		time_next.tv_sec = execute->time_sync.tv_sec;
		time_next.tv_usec = execute->time_sync.tv_usec;
		g_time_val_add(&time_next, execute->time_pos);
		
		/* How much time to next execution? */
		time_diff = irreco_time_diff(&time_current, &time_next);
		IRRECO_PRINTF("Executing next command after \"%li\" usec.\n",
			      time_diff);
			      
		{
		glong time = 2147483647;
		IRRECO_PRINTF("%li\n", time / IRRECO_SECOND_IN_USEC);
		}
			      
		/* Schedule next call. */
		if (time_diff > 0) {
			g_timeout_add(time_diff / 1000,
				      G_SOURCEFUNC(
				      irreco_cmd_chain_execute_next_1), 
				      execute);
		} else {
			g_idle_add(G_SOURCEFUNC(
				   irreco_cmd_chain_execute_next_1), 
				   execute);
		}
	
	/* End of chain. */
	} else {
		gtk_widget_destroy(execute->banner);
		g_string_free(execute->banner_message, TRUE);
		IRRECO_DEBUG("Calling call_when_done()\n");
		execute->call_when_done(execute->chain, execute->user_data);
		g_slice_free(IrrecoCmdChainExecute, execute);
	}
	
	IRRECO_RETURN_BOOL(FALSE);
}

==> irreco-core_0.6.3-4.tar.gz: irreco-core-0.6.3/src/core/irreco_cmd_chain.c <==
/*
 * irreco - Ir Remote Control
 * Copyright (C) 2007  Arto Karppinen (arto.karppinen@iki.fi)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "irreco_cmd_chain.h"
#include <hildon/hildon-banner.h>

/**
 * @addtogroup IrrecoCmdChain
 * @ingroup Irreco
 *
 * IrrecoCmdChain contains a list of IrrecoCmds, which can then all be 
 * executed one after another by calling irreco_cmd_chain_execute().
 *
 * @{ 
 */

/**
 * @file
 * Source file of @ref IrrecoCmdChain.
 */



/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/* Datatypes                                                                  */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

typedef struct _IrrecoCmdChainExecute IrrecoCmdChainExecute;
struct _IrrecoCmdChainExecute {
	
	IrrecoData     *irreco_data;
	GFunc           call_when_done;
	gpointer        user_data;
	
	GString        *banner_message;
	GtkWidget      *banner;
	
	/* Execution time will be timed in relation of time_sync. */
	GTimeVal        time_sync;
	glong		time_pos;
	
	IrrecoCmdChain *chain;
	IrrecoCmd      *command;
	guint           pos;
	guint           len;
};



/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/* Prototypes                                                                 */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

static gboolean irreco_cmd_chain_execute_next_2(IrrecoCmdChainExecute *execute);



/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/* Construction & Destruction                                                 */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

/**
 * @name Construction & Destruction 
 * @{ 
 */

IrrecoCmdChain *irreco_cmd_chain_create()
{
	IrrecoCmdChain *irreco_cmd_chain;
	IRRECO_ENTER
	
	irreco_cmd_chain = g_slice_new0(IrrecoCmdChain);
	irreco_cmd_chain->execution_rate = IRRECO_SECONDS_TO_USEC(0.3);
	IRRECO_RETURN_PTR(irreco_cmd_chain);
}

void irreco_cmd_chain_destroy(IrrecoCmdChain * irreco_cmd_chain)
{
	IRRECO_ENTER
	if (irreco_cmd_chain == NULL) IRRECO_RETURN
	irreco_cmd_chain_remove_all(irreco_cmd_chain);
	g_slice_free(IrrecoCmdChain, irreco_cmd_chain);
	IRRECO_RETURN
}

/** @} */



/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/* Private Functions                                                          */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

/**
 * @name Private Functions
 * @{ 
 */
 
/**
 * Command execution phaze one, update banner.
 */
static gboolean irreco_cmd_chain_execute_next_1(IrrecoCmdChainExecute *execute)
{
	gdouble fraction;
	IRRECO_ENTER
	
	IRRECO_PRINTF("Command %i / %u\n", execute->pos + 1, execute->len);
	execute->command = irreco_cmd_chain_get(execute->chain, execute->pos);
	
	g_string_printf(execute->banner_message, "%s%s", _("Executing: "),
			irreco_cmd_get_long_name(execute->command));
	IRRECO_PRINTF("%s\n", execute->banner_message->str);
	
	if (execute->banner == NULL) {
		GtkWidget *owner;
		owner = GTK_WIDGET(irreco_window_manager_get_gtk_window(
				   execute->irreco_data->window_manager));
		execute->banner	= hildon_banner_show_progress(
			owner, NULL, execute->banner_message->str);
	} else {
		hildon_banner_set_text(HILDON_BANNER(execute->banner), 
				       execute->banner_message->str);
	}
			       
	fraction = (float) 1 / (float) execute->len * (float)(execute->pos + 1);
	IRRECO_DEBUG("Proggressbar fraction %f\n", fraction);
	hildon_banner_set_fraction(HILDON_BANNER(execute->banner), fraction);
	g_idle_add(G_SOURCEFUNC(irreco_cmd_chain_execute_next_2), execute);
	IRRECO_RETURN_BOOL(FALSE);
}

/**
 * Command execution phaze two, execute command.
 */
static gboolean irreco_cmd_chain_execute_next_2(IrrecoCmdChainExecute *execute)
{
	gboolean result;
	IRRECO_ENTER
	
	/* Wait command delay is simply added to the time_pos. */
	if (execute->command->type == IRRECO_COMMAND_WAIT) {
		
		/* Wait command overrides the chain specific execution rate.*/
		execute->time_pos = execute->time_pos
				    + execute->command->wait.delay;
		execute->pos = execute->pos + 1;
		result = TRUE;
	} else {
		result = irreco_cmd_execute(execute->command, 
					    execute->irreco_data);
		execute->time_pos = execute->time_pos
				    + execute->chain->execution_rate;
		execute->pos = execute->pos + 1;
	}
	
	/* Is something remaining to be executed? */
	if (result == TRUE && execute->pos < execute->len) {
		
		GTimeVal time_current, time_next;
		glong time_diff;
		
		g_get_current_time(&time_current);
		time_next.tv_sec = execute->time_sync.tv_sec;
		time_next.tv_usec = execute->time_sync.tv_usec;
		g_time_val_add(&time_next, execute->time_pos);
		
		/* How much time to next execution? */
		time_diff = irreco_time_diff(&time_current, &time_next);
		IRRECO_PRINTF("Executing next command after \"%li\" usec.\n",
			      time_diff);
			      
		{
		glong time = 2147483647;
		IRRECO_PRINTF("%li\n", time / IRRECO_SECOND_IN_USEC);
		}
			      
		/* Schedule next call. */
		if (time_diff > 0) {
			g_timeout_add(time_diff / 1000,
				      G_SOURCEFUNC(
				      irreco_cmd_chain_execute_next_1), 
				      execute);
		} else {
			g_idle_add(G_SOURCEFUNC(
				   irreco_cmd_chain_execute_next_1), 
				   execute);
		}
	
	/* End of chain. */
	} else {
		gtk_widget_destroy(execute->banner);
		g_string_free(execute->banner_message, TRUE);
		IRRECO_DEBUG("Calling call_when_done()\n");
		execute->call_when_done(execute->chain, execute->user_data);
		g_slice_free(IrrecoCmdChainExecute, execute);
	}
	
	IRRECO_RETURN_BOOL(FALSE);
}

/** @} */



/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/* Public Functions                                                           */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

/**
 * @name Public Functions
 * @{ 
 */

/**
 * Remove all commands from command chain.
 */
void irreco_cmd_chain_remove_all(IrrecoCmdChain * irreco_cmd_chain)
{
	IRRECO_ENTER
	
	g_assert(irreco_cmd_chain != NULL);
	
	while (irreco_cmd_chain->command_list != NULL) {
		irreco_cmd_destroy(irreco_cmd_chain->command_list->data);
		irreco_cmd_chain->command_list = g_list_remove(
			irreco_cmd_chain->command_list,
			irreco_cmd_chain->command_list->data);
	}
	IRRECO_RETURN
}

/**
 * Set interval between the start of execution of commands inside command chain.
 */
void irreco_cmd_chain_set_execution_rate(IrrecoCmdChain * irreco_cmd_chain, 
					 glong execution_rate)
{
	IRRECO_ENTER
	g_assert(irreco_cmd_chain != NULL);
	IRRECO_PRINTF("Setting execution rate to \"%li\".\n", execution_rate);
	irreco_cmd_chain->execution_rate = execution_rate;
	IRRECO_RETURN
}

/**
 * Append a command to command chain.
 */
void irreco_cmd_chain_append(IrrecoCmdChain * irreco_cmd_chain, 
			     IrrecoCmd * irreco_cmd)
{
	IRRECO_ENTER
	
	g_assert(irreco_cmd_chain != NULL);
	
	irreco_cmd_chain->command_list = g_list_append(
		g_list_last(irreco_cmd_chain->command_list), irreco_cmd);
	IRRECO_RETURN
}

/**
 * Create a copy of command, and append the command to command chain.
 */
void irreco_cmd_chain_append_copy(IrrecoCmdChain * irreco_cmd_chain, 
				  IrrecoCmd * irreco_cmd)
{
	IRRECO_ENTER
	
	g_assert(irreco_cmd_chain != NULL);
	
	irreco_cmd_chain->command_list = g_list_append(
		g_list_last(irreco_cmd_chain->command_list),
		irreco_cmd_dublicate(irreco_cmd));
	IRRECO_RETURN
}

/**
 * Remove command by index.
 */
void irreco_cmd_chain_remove(IrrecoCmdChain * irreco_cmd_chain, guint index)
{
	IrrecoCmd *irreco_cmd;
	IRRECO_ENTER
	
	g_assert(irreco_cmd_chain != NULL);
	
	IRRECO_DEBUG("Removing command from index \"%u\".\n", index);
	irreco_cmd = (IrrecoCmd *) g_list_nth_data(g_list_first(
		irreco_cmd_chain->command_list), index);
	irreco_cmd_destroy(irreco_cmd);
	irreco_cmd_chain->command_list = g_list_remove(
		g_list_first(irreco_cmd_chain->command_list), irreco_cmd);

	IRRECO_RETURN
}

/**
 * Change order of commands inside command chain.
 */
void irreco_cmd_chain_move(IrrecoCmdChain * irreco_cmd_chain, 
			   guint from_index, guint to_index)
{
	gpointer data;
	GList* from_link;
	gint length;
	IRRECO_ENTER
	
	length = irreco_cmd_chain_length(irreco_cmd_chain);
	
	if (from_index < 0) {
		from_index = 0;
	} else if (from_index > length) {
		IRRECO_RETURN
	}
	
	if (to_index > length) {
		to_index = length;
	} else if (to_index < 0) {
		IRRECO_RETURN
	}
	
	from_link = g_list_nth(g_list_first(
		irreco_cmd_chain->command_list), from_index);
	data = from_link->data;
	irreco_cmd_chain->command_list = g_list_delete_link(
		irreco_cmd_chain->command_list, from_link);
	irreco_cmd_chain->command_list = g_list_insert(g_list_first(
		irreco_cmd_chain->command_list), data, to_index);

	IRRECO_RETURN
}

/**
 * Make a copy of a command chain.
 *
 * @param from  Make a copy of this IrrecoCmdChain, or NULL.
 */
void irreco_cmd_chain_copy(IrrecoCmdChain * from, 
			   IrrecoCmdChain * to)
{
	IRRECO_ENTER
	irreco_cmd_chain_remove_all(to);
	to->execution_rate = from->execution_rate;
	
	if (from == NULL) IRRECO_RETURN
	IRRECO_CMD_CHAIN_FOREACH(from, command)
		irreco_cmd_chain_append_copy(to, command);
	IRRECO_CMD_CHAIN_FOREACH_END

	IRRECO_RETURN
}

/**
 * Get number of command in command chain.
 */
guint irreco_cmd_chain_length(IrrecoCmdChain * irreco_cmd_chain)
{
	IRRECO_ENTER
	if (irreco_cmd_chain->command_list == NULL) IRRECO_RETURN_UINT(0);
	IRRECO_RETURN_UINT(g_list_length(g_list_first(
			   irreco_cmd_chain->command_list)));
}

/**
 * Get Nth IrrecoCmd from command chain.
 */
IrrecoCmd *irreco_cmd_chain_get(IrrecoCmdChain * irreco_cmd_chain,
				guint from_index)
{
	IRRECO_ENTER
	if (irreco_cmd_chain->command_list == NULL) IRRECO_RETURN_PTR(NULL);
	IRRECO_RETURN_PTR(g_list_nth_data(g_list_first(
			  irreco_cmd_chain->command_list), 
			  from_index));
}

/**
 * Print list of command inside command chain.
 */
void irreco_cmd_chain_print(IrrecoCmdChain * irreco_cmd_chain)
{
	gint i = 0;
	guint len;
	IRRECO_ENTER
	
	if (irreco_cmd_chain == NULL) {
		IRRECO_ERROR("IrrecoCmdChain pointer is NULL.\n");
		IRRECO_RETURN
	}
	
	len = irreco_cmd_chain_length(irreco_cmd_chain);
	IRRECO_PRINTF("Printing command chain. Lenght: %u\n", len);
	IRRECO_CMD_CHAIN_FOREACH(irreco_cmd_chain, command)
		IRRECO_PRINTF("Command %i / %u\n", ++i, len);
		irreco_cmd_print(command);
	IRRECO_CMD_CHAIN_FOREACH_END

	IRRECO_RETURN
}

/**
 * Start executing the command chain.
 *
 * If the command chain is longer than one command, then delayed execution is 
 * started. In that case, call_when_done will be called when the command chain 
 * has finished executing.
 *
 * Data pointer will be the command chain, and user_data what user has given.
 *
 * @return TRUE if delayed execution started, FALSE otherwise.
 */
gboolean irreco_cmd_chain_execute(IrrecoCmdChain * irreco_cmd_chain,
				  IrrecoData *irreco_data,
				  GFunc call_when_done,
				  gpointer user_data)
{
	guint len;
	IRRECO_ENTER
	
	/* Empty list. */
	if (irreco_cmd_chain->command_list == NULL) {
		IRRECO_RETURN_BOOL(FALSE);
		
	/* Only one comamnd. */
	} else if ((len = irreco_cmd_chain_length(irreco_cmd_chain)) == 1) {
		irreco_cmd_execute(irreco_cmd_chain_get(irreco_cmd_chain, 0),
				   irreco_data);
		IRRECO_RETURN_BOOL(FALSE);
			
	/* Multiple commands. */
	} else {
		IrrecoCmdChainExecute *execute;
		
		execute = g_slice_new0(IrrecoCmdChainExecute);
		execute->banner_message = g_string_new("");
		execute->irreco_data = irreco_data;
		execute->chain       = irreco_cmd_chain;
		execute->pos         = 0;
		execute->len         = len;
		
		execute->call_when_done	= call_when_done;
		execute->user_data = user_data;
		
		g_get_current_time(&execute->time_sync);
		irreco_cmd_chain_execute_next_1(execute);
	}
	IRRECO_RETURN_BOOL(TRUE);
}

/** @} */
